/*
 * modelnet DumpManager.cc
 *
 *    Used for managing tcpdump dumping
 *
 * Copyright (c) 2006 Calvin Hubble
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *    * Neither the names of Duke University nor The University of
 *      California, San Diego, nor the names of the authors or contributors
 *      may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */


#include <stdio.h>
// #include <hash_set>
#include <string>
#include <sys/time.h>
#include <iostream>
#include "EdgeInfo.h"
#include "DumpManager.h"
#include "savefile.h"

using namespace std;

DumpManager::DumpManager()
{
}


bool DumpManager::addEdge(EdgeInfo edgeInfo)
{
  edgeIdMap[edgeInfo.getId()] = edgeInfo;
  edgeFPMap[edgeInfo.getId()] = createNewDumpFile(edgeInfo.getDumpLocation());
  
  if (edgeFPMap[edgeInfo.getId()] == NULL)
  {
    return false;
  }

  return true;
}

// HandleModlenetData
// Given data from modelnet, will handl dumping the packet
// and return the amount of data consumed, or -1 on an error

int DumpManager::handleModelnetData(unsigned char* data, int maxReadAmount)
{
  unsigned char* finalData;
  unsigned char srcMAC[6];
  unsigned char dstMAC[6];
  int consumedData = 0;

  // Check if we have enough space to read the header
  if (maxReadAmount < sizeof(struct modelnet_hdr))
    return 0;
  
  // Pull off modelnet header
  struct modelnet_hdr* mn_hdr;
  mn_hdr = (struct modelnet_hdr*)data;
  data += sizeof(struct modelnet_hdr);
  consumedData += sizeof(struct modelnet_hdr);

//   cout << "Handling data for link " << mn_hdr->linkId << endl;
//   cout << "pktLength: " << mn_hdr->pktLength <<endl;;
//   cout << "captureLength: " << mn_hdr->capturedLength << endl << endl;  
  // Get information about which edge we are dumping
  EdgeInfo edgeInfo = edgeIdMap[mn_hdr->linkId];

  // Create MAC addresses from src & destination IDs
  memset(&srcMAC, 0, 6);
  memset(&dstMAC, 0, 6);
  *((int*)(srcMAC)) = edgeInfo.getSrc();
  *((int*)(dstMAC)) = edgeInfo.getDst();
  
  // Copy data into pcap header
  struct pcap_pkthdr pkt_hdr;
  memcpy(&pkt_hdr, &mn_hdr->tv, sizeof(struct timeval));
  pkt_hdr.len = mn_hdr->pktLength + 14;
  pkt_hdr.caplen = mn_hdr->capturedLength + 14;
  
  // Make sure we have enough buffer to copy from
  if (maxReadAmount < 
      (sizeof(struct modelnet_hdr) + mn_hdr->capturedLength))
    return 0;
  
  // Now we can make our data packet
  // adding ethernet header on first
  finalData = (unsigned char*) malloc(mn_hdr->capturedLength + 14);  
  memcpy(finalData, dstMAC, 6);
  memcpy(finalData+6, srcMAC, 12);
  finalData[12] = 0x08; // MAC magic
  finalData[13] = 0x00; // MAC magic
  memcpy(finalData+14, data, mn_hdr->capturedLength);
  consumedData += mn_hdr->capturedLength;

  // Get our file pointer
  FILE* fp = getLinkFP(mn_hdr->linkId);
  if (fp == NULL)
  {
    free(finalData);
    cerr << "Bad file pointer for link ID: " << mn_hdr->linkId << endl;
    return -1;
  }

  pcap_dump((unsigned char*)fp, &pkt_hdr, (unsigned char*)finalData);
  free(finalData);
  return consumedData;
}

FILE* DumpManager::createNewDumpFile(string fileName)
{
  FILE* fp = fopen(fileName.c_str(), "wb");
  if (fp == NULL)
  {
    cerr << "Error opening file " << fileName;
    return NULL;
  }

  /* XX - hardcoded values for linktype as well as snapshot length */
  /* link types defined in /usr/include/net/bpf.h */
  if (sf_write_header(fp, DLT_EN10MB, 0, TCPDUMP_BUF_SIZE) == -1)
  {
    fclose(fp);
    return NULL;
  }

  return fp;
}

FILE* DumpManager::getLinkFP(int linkId)
{
  return edgeFPMap[linkId];
}

void DumpManager::closeAllFiles()
{
  
}

void DumpManager::flushAllFiles()
{
  EdgeFPMap::const_iterator i;

  for (i = edgeFPMap.begin(); i != edgeFPMap.end(); i++)
  {
    fflush(i->second);
  }
}

DumpManager::~DumpManager()
{
  closeAllFiles();
}
